// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccaudio.
// 
// The exception is that, if you link the ccaudio library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccaudio library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccaudio.  If you copy code from other releases into a copy of
// ccaudio, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccaudio, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include <cc++/strchar.h>
#include "private.h"
#include "audio.h"

#ifdef	CCXX_NAMESPACES
namespace ost {
#endif

void AudioFile::setaushort(unsigned char *data, unsigned short val)
{
	if(info.order == __BIG_ENDIAN)
	{
		data[0] = val / 256;
		data[1] = val % 256;
	}
	else
	{
		data[1] = val / 256;
		data[0] = val % 256;
	}
}

unsigned short AudioFile::getaushort(unsigned char *data)
{
	if(info.order == __BIG_ENDIAN)
		return data[0] * 256 + data[1];
	else
		return data[1] * 256 + data[0];
}	

void AudioFile::setaulong(unsigned char *data, unsigned long val)
{
	int i = 4;

	while(i-- > 0)
	{
		if(info.order == __BIG_ENDIAN)
			data[i] = val & 0xff;
		else
			data[3 - i] = val & 0xff;
		val /= 256;
	}
}

unsigned long AudioFile::getaulong(unsigned char * data)
{
	int i = 4;
	unsigned long val =0;
	
	while(i-- > 0)
	{
		if(info.order == __BIG_ENDIAN)
			val = (val << 8) | data[3 - i];
		else
			val = (val << 8) | data[i];
	}
	return val;
}

AudioFile::AudioFile(const char *name, unsigned long sample)
{
	Initialize();
	Open(name);
	if(!isOpen())
		return;
	setPosition(sample);
}

AudioFile::AudioFile(const char *name, audioinfo_t *info, unsigned long samples)
{
	Initialize();
	Create(name, info);
	if(!isOpen())
		return;
	setMinimum(samples);
}

void AudioFile::Create(const char *name, audioinfo_t *myinfo)
{
	unsigned char aufile[24];
	unsigned char riffhdr[40];
	unsigned int count;

	if(!afCreate(name))
		return;

	memcpy(&info, myinfo, sizeof(info));
	info.annotation = NULL;
	pathname = new char[strlen(name) + 1];
	strcpy(pathname, name);
	if(myinfo->annotation)
	{
		info.annotation = new char[strlen(myinfo->annotation) + 1];
		strcpy(info.annotation, myinfo->annotation);
	}
	
	switch(info.format)
	{
	case AUDIO_FORMAT_RIFF:
		if(!info.order)
			info.order = __LITTLE_ENDIAN;
		if(info.order == __LITTLE_ENDIAN)
			strncpy((char *)riffhdr, "RIFF", 4);
		else
			strncpy((char *)riffhdr, "RIFX", 4);
		if(!info.rate)
			info.rate = samplerate(info.encoding);

		header = 46;

		memset(riffhdr + 4, 0xff, 4);
		strncpy((char *)riffhdr + 8, "WAVE", 4);
		if(afWrite(riffhdr, 12) != 12)
		{
			Close();
			return;
		}

		strncpy((char *)riffhdr, "fmt ", 4);
		setaulong(riffhdr + 4, 18);
		setaushort(riffhdr + 8, 1);
		if(ismono(info.encoding))
			setaushort(riffhdr + 10, 1);
		else
			setaushort(riffhdr + 10, 2);
		setaulong(riffhdr + 12, info.rate);
		setaulong(riffhdr + 16, tobytes(info.encoding, info.rate));
		setaulong(riffhdr + 20, tobytes(info.encoding, info.rate) / 8);
		setaushort(riffhdr + 24, 0);

		switch(info.encoding)
		{
		case PCM8_AUDIO_ENCODING:
		case PCM8_STEREO_ENCODING:
			setaushort(riffhdr + 24, 8);
			break;
		case PCM16_AUDIO_ENCODING:
		case PCM16_STEREO_ENCODING:
		case CDA_MONO_ENCODING:
		case CDA_STEREO_ENCODING:
			setaushort(riffhdr + 24, 16);
			break;
		case PCM32_AUDIO_ENCODING:
		case PCM32_STEREO_ENCODING:
			setaushort(riffhdr + 24, 32);
			break;
		case ALAW_AUDIO_ENCODING:
			setaushort(riffhdr + 8, 6);
			break;
		case MULAW_AUDIO_ENCODING:
			setaushort(riffhdr + 8, 7);
			break;
		case OKI_ADPCM_ENCODING:
			setaushort(riffhdr + 8, 0x10);
			break;
		case DIALOGIC_ADPCM_ENCODING:
			setaushort(riffhdr + 8, 0x17);
			break;
		case G721_ADPCM_ENCODING:
			setaushort(riffhdr + 8, 0x40);
			break;
		case G722_AUDIO_ENCODING:
			setaushort(riffhdr + 8, 0x64);
			break;
		case GSM_VOICE_ENCODING:
			setaushort(riffhdr + 8, 0x31);
			break;
		case G723_3BIT_ENCODING:
		case G723_5BIT_ENCODING:
			setaushort(riffhdr + 8, 0x14);
		}
		if(afWrite(riffhdr, 26) != 26)
		{
			Close();
			return;
		}

		strncpy((char *)riffhdr, "data", 4);
		memset(riffhdr, 0xff, 4);
		if(afWrite(riffhdr, 8) != 8)
		{
			Close();
			return;
		}
		break;

	case AUDIO_FORMAT_SUN:
		if(!info.order)
			info.order = __BIG_ENDIAN;

		if(!info.rate)
			info.rate = samplerate(info.encoding);

		strncpy((char *)aufile, ".snd", 4);
		setaulong(aufile + 4, 24 + strlen(info.annotation) + 1);
		header = getaulong(aufile + 4);
		setaulong(aufile + 8, ~0l);
		switch(info.encoding)
		{
		case PCM8_STEREO_ENCODING:
		case PCM8_AUDIO_ENCODING:
			setaulong(aufile + 12, 2);
			break;
		case PCM16_STEREO_ENCODING:
		case PCM16_AUDIO_ENCODING:
		case CDA_STEREO_ENCODING:
		case CDA_MONO_ENCODING:
			setaulong(aufile + 12, 3);
			break;
		case PCM32_STEREO_ENCODING:
		case PCM32_AUDIO_ENCODING:
			setaulong(aufile + 12, 5);
			break;
		case G721_ADPCM_ENCODING:
			setaulong(aufile + 12, 23);
			break;
		case G722_AUDIO_ENCODING:
		case G722_7BIT_ENCODING:
		case G722_6BIT_ENCODING:
			setaulong(aufile + 12, 24);
			break;
		case G723_3BIT_ENCODING:
			setaulong(aufile + 12, 25);
			break;
		case G723_5BIT_ENCODING:
			setaulong(aufile + 12, 26);
			break;
		case ALAW_AUDIO_ENCODING:
			setaulong(aufile + 12, 27);
			break;
		default:
			setaulong(aufile + 12, 1);
		}
		setaulong(aufile + 16, info.rate);
		if(ismono(info.encoding))
			setaulong(aufile + 20, 1);
		else
			setaulong(aufile + 20, 2);
		if(afWrite(aufile, 24) != 24)
		{
			Close();
			return;
		}
		if(info.annotation)
			afWrite((unsigned char *)info.annotation, 
				strlen(info.annotation) + 1);
	}
}

void AudioFile::getWaveFormat(int request)
{
	unsigned char filehdr[24];
	unsigned int count;
	int bitsize;
	int channels;

	if(request > 24)
		request = 24;

	if(!afPeek(filehdr, request))
	{
		Close();
		return;
	}
	channels = getaushort(filehdr + 2);
	info.rate = getaulong(filehdr + 4);

	switch(getaushort(filehdr))
	{
	case 1:
		bitsize = getaushort(filehdr + 14);
		switch(bitsize)
		{
		case 8:
			if(channels > 1)
				info.encoding = PCM8_STEREO_ENCODING;
			else
				info.encoding = PCM8_AUDIO_ENCODING;
			break;
		case 16:
			if(info.rate == 44100)
			{
				if(channels > 1)
					info.encoding = CDA_STEREO_ENCODING;
				else
					info.encoding = CDA_MONO_ENCODING;
				break;
			}
			if(channels > 1)
					info.encoding = PCM16_STEREO_ENCODING;
			else
				info.encoding = PCM16_AUDIO_ENCODING;
			break;
		case 32:
			if(channels > 1)
				info.encoding = PCM32_STEREO_ENCODING;
			else
				info.encoding = PCM32_AUDIO_ENCODING;
			break;
		default:
			info.encoding = UNKNOWN_AUDIO_ENCODING;
		}
		break;
	case 6:
		info.encoding = ALAW_AUDIO_ENCODING;
		break;
	case 7:
		info.encoding = MULAW_AUDIO_ENCODING;
		break;
	case 0x10:
		info.encoding = OKI_ADPCM_ENCODING;
		break;
	case 0x17:
		info.encoding = DIALOGIC_ADPCM_ENCODING;
		break;
	case 0x40:
		info.encoding = G721_ADPCM_ENCODING;
		break;
	case 0x65:
		info.encoding = G722_AUDIO_ENCODING;
		break;
	case 0x31:
		info.encoding = GSM_VOICE_ENCODING;
		break;
	case 0x14:
		bitsize = getaulong(filehdr + 8) * 8 / info.rate;
		if(bitsize == 3)
			info.encoding = G723_3BIT_ENCODING;
		else
			info.encoding = G723_5BIT_ENCODING;
		break; 
	default:
		info.encoding = UNKNOWN_AUDIO_ENCODING;
	}
}
	 
void AudioFile::Open(const char *name)
{
	unsigned char filehdr[24];
	unsigned int count;
	unsigned bitsize;
	char *ext;
	unsigned channels;

	if(!afOpen(name))
		return;

	pathname = new char[strlen(name) + 1];
	strcpy(pathname, name);
	header = 0l;

	info.encoding = MULAW_AUDIO_ENCODING;
	info.format = AUDIO_FORMAT_RAW;
	info.order = 0;
	ext = strrchr(pathname, '.');
	if(!ext)
		return;

	if(!stricmp(ext, ".ul"))
	{
		info.encoding = MULAW_AUDIO_ENCODING;
		return;
	}

	if(!stricmp(ext, ".al"))
	{
		info.encoding = ALAW_AUDIO_ENCODING;
		return;
	}

	if(!stricmp(ext, ".vox"))
	{
		info.encoding = DIALOGIC_ADPCM_ENCODING;
		return;
	}

	if(!stricmp(ext, ".adpcm"))
	{
		info.encoding = G721_ADPCM_ENCODING;
		return;
	}

	if(!stricmp(ext, ".a24"))
	{
		info.encoding = G723_3BIT_ENCODING;
		return;
	}

	if(!stricmp(ext, ".a40"))
	{
		info.encoding = G723_5BIT_ENCODING;
		return;
	}

	strcpy((char *)filehdr, ".xxx");

	if(!afPeek(filehdr, 24))
	{
		Close();
		return;
	}

	if(!strncmp((char *)filehdr, "RIFF", 4))
	{
		info.format = AUDIO_FORMAT_RIFF;
		info.order = __LITTLE_ENDIAN;
	}

	if(!strncmp((char *)filehdr, "RIFX", 4))
	{
		info.order = __BIG_ENDIAN;
		info.format = AUDIO_FORMAT_RIFF;
	}
		
	if(!strncmp((char *)filehdr + 8, "WAVE", 4) && 
	   info.format == AUDIO_FORMAT_RIFF)
	{
		header = 12;
		for(;;)
		{
			if(!afSeek(header))
			{
				Close();
				return;
			}
			if(!afPeek(filehdr, 8))
			{
				Close();
				return;
			}
			header += 8;
			if(!strncmp((char *)filehdr, "data", 4))
				break;

			count = getaulong(filehdr + 4);
			header += count;
			if(!strncmp((char *)filehdr, "fmt ", 4))
				getWaveFormat(count);		
			
		}
		afSeek(header);
		return;
	}
	if(!strncmp((char *)filehdr, ".snd", 4))
	{
		info.format = AUDIO_FORMAT_SUN;
		info.order = __BIG_ENDIAN;
		header = getaulong(filehdr + 4);
		info.rate = getaulong(filehdr + 16);
		channels = getaulong(filehdr + 20);

		switch(getaulong(filehdr + 12))
		{
		case 2:
			if(channels > 1)
				info.encoding = PCM8_STEREO_ENCODING;
			else
				info.encoding = PCM8_AUDIO_ENCODING;
			break;
		case 3:
			if(info.rate == 44100)
			{
				if(channels > 1)
					info.encoding = CDA_STEREO_ENCODING;
				else
					info.encoding = CDA_MONO_ENCODING;
				break;
			}
			if(channels > 1)
				info.encoding = PCM16_STEREO_ENCODING;
			else
				info.encoding = PCM16_AUDIO_ENCODING;
			break;	
		case 5:
			if(channels > 1)
				info.encoding = PCM32_STEREO_ENCODING;
			else
				info.encoding = PCM32_AUDIO_ENCODING;
			break;
		case 23:
			info.encoding = G721_ADPCM_ENCODING;
			break;
		case 24:
			info.encoding = G722_AUDIO_ENCODING;
			break;
		case 25:
			info.encoding = G723_3BIT_ENCODING;
			break;
		case 26:
			info.encoding = G723_5BIT_ENCODING;
			break;
		case 27:
			info.encoding = ALAW_AUDIO_ENCODING;
			break;
		case 1:
			info.encoding = MULAW_AUDIO_ENCODING;
			break;
		default:
			info.encoding = UNKNOWN_AUDIO_ENCODING;
		}
		if(header > 24)
		{
			info.annotation = new char[header - 24];
			afSeek(24);
			afRead((unsigned char *)info.annotation, header - 24);
		}
		return;
	}
	else
		afSeek(0);
}
 

void AudioFile::Close(void)
{
	afClose();
	if(pathname)
	{
		delete[] pathname;
		pathname = NULL;
	}
	if(info.annotation)
	{
		delete[] info.annotation;
		info.annotation = NULL;
	}
	minimum = 0;
}

audioerror_t AudioFile::getInfo(audioinfo_t *infobuf)
{
	if(!isOpen())
		return setError(AUDIO_NOT_OPENED);

	if(!infobuf)
		return setError(AUDIO_REQUEST_INVALID);

	memcpy(infobuf, &info, sizeof(info));
	return AUDIO_SUCCESS;
}

audioerror_t AudioFile::setError(audioerror_t err)
{
	if(err)
		error = err;	
	return err;
}

audioerror_t AudioFile::setMinimum(unsigned long samples)
{
	if(!isOpen())
		return setError(AUDIO_NOT_OPENED);
	minimum = samples;
	return AUDIO_SUCCESS;
}

int AudioFile::getBuffer(void *addr, unsigned bytes)
{
	char *fname;
	unsigned char *caddr = (unsigned char *)addr;
	int count, xfer = 0;
	for(;;)
	{
		count = afRead(caddr, bytes);
		if(count < 0)
		{
			if(!xfer)
				return count;
			return xfer;
		}
		xfer += count;
		if(count == bytes)
			return xfer;
		fname = getContinuation();
		if(!fname)
			return xfer;
		Close();
		Open(fname);
		if(!isOpen())
			return xfer;
		bytes -= count;
		caddr += count;
	}
}	

audioerror_t AudioFile::getSamples(void *addr, unsigned request)
{
	char *fname;
	unsigned char *caddr = (unsigned char *)addr;
	int count, bytes;
	for(;;)
	{
		int bytes = tobytes(info.encoding, request);
		if(bytes < 1)
			return setError(AUDIO_REQUEST_INVALID);
		count = afRead(caddr, bytes);
		if(count == bytes)
			return AUDIO_SUCCESS;

		if(count < 0)
			return AUDIO_READ_FAILURE;
	
		if(count > 0)
		{
			caddr += tobytes(info.encoding, count);
			count = request - tosamples(info.encoding, count);
		}
		else
			count = request;

		fname = getContinuation();
		if(!fname)
			break;

		Close();
		Open(fname);
		if(!isOpen())
			break;

		request = count;
	}
	if(count)
		samplefill(caddr, count, info.encoding);
	return AUDIO_READ_INCOMPLETE;
}

int AudioFile::putBuffer(void *addr, unsigned len)
{
	return afWrite((unsigned char *)addr, len);
}

audioerror_t AudioFile::putSamples(void *addr, unsigned samples)
{
	int count;
	int bytes = tobytes(info.encoding, samples);
	if(bytes < 1)
		return setError(AUDIO_REQUEST_INVALID);

	count = afWrite((unsigned char *)addr, bytes);
	if(count == bytes)
		return AUDIO_SUCCESS;

	if(count < 1)
		return AUDIO_WRITE_FAILURE;
	else
		return AUDIO_WRITE_INCOMPLETE;
}		

audioerror_t AudioFile::Skip(long samples)
{
	long pos = getPosition() + samples;
	if(pos < 0)
		pos = 0;

	setPosition(pos);
	return AUDIO_SUCCESS;
}

#ifdef	CCXX_NAMESPACES
};
#endif


