// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccaudio.
// 
// The exception is that, if you link the ccaudio library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccaudio library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccaudio.  If you copy code from other releases into a copy of
// ccaudio, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccaudio, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include "private.h"
#include "audio.h"

#ifdef	CCXX_NAMESPACES
namespace ost {
#endif

int samplecount(audioencoding_t encoding)
{
	switch(encoding)
	{
	case GSM_VOICE_ENCODING:
		return 160;
	case UNKNOWN_AUDIO_ENCODING:
		return 0;
	case G723_3BIT_ENCODING:
	case G723_5BIT_ENCODING:
		return 8;
	case G721_ADPCM_ENCODING:
	case OKI_ADPCM_ENCODING:
	case DIALOGIC_ADPCM_ENCODING:
		return 2;
	default:
		return 1;
	}
}

int sampleframe(audioencoding_t encoding, int samples)
{
	int framing = 0;
	switch(encoding)
	{
	case GSM_VOICE_ENCODING:
		framing = 33;
		break;
	case G723_3BIT_ENCODING:
		framing = 3;
		break;
	case G723_5BIT_ENCODING:
		framing = 5;
		break;
	case UNKNOWN_AUDIO_ENCODING:
		return 0;
	case PCM32_STEREO_ENCODING:
		return 8;
	case PCM32_AUDIO_ENCODING:
	case PCM16_STEREO_ENCODING:
	case CDA_STEREO_ENCODING:
		framing = 4;
		break;
	case PCM8_STEREO_ENCODING:
	case PCM16_AUDIO_ENCODING:
	case CDA_MONO_ENCODING:
		framing = 2;
		break;
	default:
		framing = 1;
	}
	if(!samples)
		return framing;

	return (samples / framing) * framing;
}

void samplefill(unsigned char *addr, int samples, audioencoding_t encoding)
{
	int frame = sampleframe(encoding);
	int count = samplecount(encoding);
	
	if(!frame || !count)
		return;

	while(samples >= count)
	{
		switch(encoding)
		{
		case MULAW_AUDIO_ENCODING:
			*addr = 0xff;
			break;
		case ALAW_AUDIO_ENCODING:
			*addr = 0x55;
			break;
		default:
			memset(addr, 0, frame);
			break;
		}
		addr += frame;
		samples -= count;
	}
}

samplerate_t samplerate(audioencoding_t encoding)
{
	switch(encoding)
	{
	case PCM8_STEREO_ENCODING:
	case PCM8_AUDIO_ENCODING:
	case PCM16_STEREO_ENCODING:
	case PCM16_AUDIO_ENCODING:
	case PCM32_STEREO_ENCODING:
	case PCM32_AUDIO_ENCODING:
	case UNKNOWN_AUDIO_ENCODING:
		return SAMPLE_RATE_UNKNOWN;
	case DIALOGIC_ADPCM_ENCODING:
		return SAMPLE_RATE_6KHZ;
	case CDA_STEREO_ENCODING:
	case CDA_MONO_ENCODING:
		return SAMPLE_RATE_44KHZ;
	}
	return SAMPLE_RATE_8KHZ;
}

unsigned long tosamples(audioencoding_t encoding, size_t bytes)
{
	unsigned long sf = sampleframe(encoding);
	if(!bytes || !sf)
		return 0;
	unsigned long frames = bytes / sf;
	return frames * samplecount(encoding);
}

unsigned long tobytes(audioencoding_t encoding, unsigned long samples)
{
	unsigned long sc = samplecount(encoding);
	if(!samples || !sc)
		return 0;
	unsigned long frames = samples / sc;
	return frames * sampleframe(encoding);
}

bool ismono(audioencoding_t encoding)
{
	switch(encoding)
	{
	case PCM8_STEREO_ENCODING:
	case PCM16_STEREO_ENCODING:
	case PCM32_STEREO_ENCODING:
	case CDA_STEREO_ENCODING:
		return false;
	}
	return true;
}

bool issterio(audioencoding_t encoding)
{
	return !ismono(encoding);
}

#ifdef	CCXX_NAMESPACES
};
#endif

